@@ -101,12 +101,13 @@
 //! }
 //! ```
 //!
-//! - Change the return type of the delegated method using a `From` or `TryFrom` impl or omit it altogether
+//! - Use modifiers that alter the generated method body
 //! ```rust
 //! use delegate::delegate;
 //! struct Inner;
 //! impl Inner {
 //!     pub fn method(&self, num: u32) -> u32 { num }
+//!     pub fn method_res(&self, num: u32) -> Result<u32, ()> { Ok(num) }
 //! }
 //! struct Wrapper { inner: Inner }
 //! impl Wrapper {
@@ -125,14 +126,52 @@
 //!             #[call(method)]
 //!             pub fn method2(&self, num: u32) -> Result<u16, std::num::TryFromIntError>;
 //!
-//!             // calls method, converts result to i6 using `TryFrom`, unwrap the result
-//!             #[try_into(unwrap)]
+//!             // calls method_res, unwraps the result
+//!             #[unwrap]
+//!             pub fn method_res(&self, num: u32) -> u32;
+//!
+//!             // calls method_res, unwraps the result, then calls into
+//!             #[unwrap]
+//!             #[into]
+//!             #[call(method_res)]
+//!             pub fn method_res_into(&self, num: u32) -> u64;
+//!
+//!             // specify explicit type for into
+//!             #[into(u64)]
 //!             #[call(method)]
-//!             pub fn method3(&self, num: u32) -> u16;
+//!             pub fn method_into_explicit(&self, num: u32) -> u64;
 //!         }
 //!     }
 //! }
 //! ```
+//!
+//! - Custom called expression
+//!
+//! The `#[expr()]` attribute can be used to modify the delegated call. You can use the `$` sigil as a placeholder for what delegate would normally expand to, and wrap that expression with custom code.
+//!
+//! _Note:_ the `$` placeholder isn't required and can be present multiple times if you want.
+//!
+//! ```rs
+//! struct A(Vec<u8>);
+//!
+//! impl A {
+//!     delegate! {
+//!         to self.0 {
+//!             #[expr(*$.unwrap())]
+//!             /// Here `$` == `self.0.get(idx)`
+//!             /// Will expand to `*self.0.get(idx).unwrap()`
+//!             fn get(&self, idx: usize) -> u8;
+//!
+//!             #[call(get)]
+//!             #[expr($?.checked_pow(2))]
+//!             /// Here `$` == `self.0.get(idx)`
+//!             /// Will expand to `self.0.get(idx)?.checked_pow(2)`
+//!             fn get_checked_pow_2(&self, idx: usize) -> Option<u8>;
+//!         }
+//!     }
+//! }
+//! ```
+//!
 //! - Call `await` on async functions
 //! ```rust
 //! use delegate::delegate;
@@ -183,6 +222,30 @@
 //! }
 //! ```
 //! - Inserts `#[inline(always)]` automatically (unless you specify `#[inline]` manually on the method)
+//! - You can use an attribute on a whole segment to automatically apply it to all methods in that
+//!   segment:
+//! ```rust
+//! use delegate::delegate;
+//!
+//! struct Inner;
+//!
+//! impl Inner {
+//!   fn foo(&self) -> Result<u32, ()> { Ok(0) }
+//!   fn bar(&self) -> Result<u32, ()> { Ok(1) }
+//! }
+//!
+//! struct Wrapper { inner: Inner }
+//!
+//! impl Wrapper {
+//!   delegate! {
+//!     #[unwrap]
+//!     to self.inner {
+//!       fn foo(&self) -> u32; // calls self.inner.foo().unwrap()
+//!       fn bar(&self) -> u32; // calls self.inner.bar().unwrap()
+//!     }
+//!   }
+//! }
+//! ```
 //! - Specify expressions in the signature that will be used as delegated arguments
 //! ```rust
 //! use delegate::delegate;
@@ -210,9 +273,10 @@
 //! }
 //! ```
 //! - Modify how will an input parameter be passed to the delegated method with parameter attribute modifiers.
-//! Currently, the following modifiers are supported:
+//!   Currently, the following modifiers are supported:
 //!     - `#[into]`: Calls `.into()` on the parameter passed to the delegated method.
 //!     - `#[as_ref]`: Calls `.as_ref()` on the parameter passed to the delegated method.
+//!     - `#[newtype]`: Calls `.0` on the parameter passed to the delegated method.
 //! ```rust
 //! use delegate::delegate;
 //!
@@ -237,18 +301,197 @@
 //!     }
 //! }
 //! ```
+//! - Specify a trait through which will the delegated method be called
+//!   (using [UFCS](https://doc.rust-lang.org/reference/expressions/call-expr.html#disambiguating-function-calls).
+//! ```rust
+//! use delegate::delegate;
+//!
+//! struct InnerType {}
+//! impl InnerType {
+//!     
+//! }
+//!
+//! trait MyTrait {
+//!   fn foo(&self);
+//! }
+//! impl MyTrait for InnerType {
+//!   fn foo(&self) {}
+//! }
+//!
+//! struct Wrapper(InnerType);
+//! impl Wrapper {
+//!     delegate! {
+//!         to &self.0 {
+//!             // Calls `MyTrait::foo(&self.0)`
+//!             #[through(MyTrait)]
+//!             pub fn foo(&self);
+//!         }
+//!     }
+//! }
+//! ```
+//!
+//! - Add additional arguments to method
+//!
+//!  ```rust
+//!  use delegate::delegate;
+//!  use std::cell::OnceCell;
+//!  struct Inner(u32);
+//!  impl Inner {
+//!      pub fn new(m: u32) -> Self {
+//!          // some "very complex" constructing work
+//!          Self(m)
+//!      }
+//!      pub fn method(&self, n: u32) -> u32 {
+//!          self.0 + n
+//!      }
+//!  }
+//!  
+//!  struct Wrapper {
+//!      inner: OnceCell<Inner>,
+//!  }
+//!  
+//!  impl Wrapper {
+//!      pub fn new() -> Self {
+//!          Self {
+//!              inner: OnceCell::new(),
+//!          }
+//!      }
+//!      fn content(&self, val: u32) -> &Inner {
+//!          self.inner.get_or_init(|| Inner(val))
+//!      }
+//!      delegate! {
+//!          to |k: u32| self.content(k) {
+//!              // `wrapper.method(k, num)` will call `self.content(k).method(num)`
+//!              pub fn method(&self, num: u32) -> u32;
+//!          }
+//!      }
+//!  }
+//!  ```
+//! - Delegate associated functions
+//!   ```rust
+//!   use delegate::delegate;
+//!
+//!   struct A {}
+//!   impl A {
+//!       fn foo(a: u32) -> u32 {
+//!           a + 1
+//!       }
+//!   }
+//!
+//!   struct B;
+//!
+//!   impl B {
+//!       delegate! {
+//!           to A {
+//!               fn foo(a: u32) -> u32;
+//!           }
+//!       }
+//!   }
+//!
+//!   assert_eq!(B::foo(1), 2);
+//!   ```
+//! - Delegate associated constants
+//!
+//! ```rust
+//! use delegate::delegate;
+//!
+//! trait WithConst {
+//!     const TOTO: u8;
+//! }
+//!
+//! struct A;
+//! impl WithConst for A {
+//!     const TOTO: u8 = 1;
+//! }
+//!
+//! struct B;
+//! impl WithConst for B {
+//!     const TOTO: u8 = 2;
+//! }
+//! struct C;
+//! impl WithConst for C {
+//!     const TOTO: u8 = 2;
+//! }
+//!
+//! enum Enum {
+//!     A(A),
+//!     B(B),
+//!     C(C),
+//! }
+//!
+//! impl Enum {
+//!     delegate! {
+//!         to match self {
+//!             Self::A(a) => a,
+//!             Self::B(b) => b,
+//!             Self::C(c) => { println!("hello from c"); c },
+//!         } {
+//!             #[const(WithConst::TOTO)]
+//!             fn get_toto(&self) -> u8;
+//!         }
+//!     }
+//! }
+//!
+//! assert_eq!(Enum::A(A).get_toto(), <A as WithConst>::TOTO);
+//! ```
+//!
+//! - Delegate to fields
+//! ```rust
+//! use delegate::delegate;
+//!
+//! struct Datum {
+//!     value: u32,
+//!     error: u32,
+//! }
+//!
+//! struct DatumWrapper(Datum);
+//!
+//! impl DatumWrapper {
+//!     delegate! {
+//!         to self.0 {
+//!             /// Get the value of a nested field with the same name
+//!             #[field]
+//!             fn value(&self) -> u32;
+//!
+//!             /// Get the value of a nested field with a different name
+//!             #[field(value)]
+//!             fn renamed_value(&self) -> u32;
+//!
+//!             /// Get shared reference to a nested field
+//!             #[field(&value)]
+//!             fn value_ref(&self) -> &u32;
+//!
+//!             /// Get mutable reference to a nested field
+//!             #[field(&mut value)]
+//!             fn value_ref_mut(&mut self) -> &mut u32;
+//!
+//!             /// Get mutable reference to a nested field with the same name
+//!             #[field(&)]
+//!             fn error(&self) -> &u32;
+//!         }
+//!     }
+//! }
+//! ```
 
 extern crate proc_macro;
+use std::mem;
 
+use attributes::AssociatedConstant;
 use proc_macro::TokenStream;
-use proc_macro2::Ident;
 
+use proc_macro2::Ident;
 use quote::{quote, ToTokens};
-use std::collections::HashMap;
-use syn::parse::ParseStream;
+use syn::parse::{Parse, ParseStream};
 use syn::spanned::Spanned;
 use syn::visit_mut::VisitMut;
-use syn::{parse_quote, Attribute, Error, ExprMethodCall, Meta};
+use syn::{parse_quote, Error, Expr, ExprField, ExprMethodCall, FnArg, GenericParam, Meta};
+
+use crate::attributes::{
+    combine_attributes, parse_method_attributes, parse_segment_attributes, ReturnExpression,
+    SegmentAttributes, TargetSpecifier,
+};
+
+mod attributes;
 
 mod kw {
     syn::custom_keyword!(to);
@@ -259,6 +502,7 @@ mod kw {
 enum ArgumentModifier {
     Into,
     AsRef,
+    Newtype,
 }
 
 #[derive(Clone)]
@@ -271,8 +515,7 @@ enum DelegatedInput {
 }
 
 fn get_argument_modifier(attribute: syn::Attribute) -> Result<ArgumentModifier, Error> {
-    let meta = attribute.parse_meta()?;
-    if let Meta::Path(mut path) = meta {
+    if let Meta::Path(mut path) = attribute.meta {
         if path.segments.len() == 1 {
             let segment = path.segments.pop().unwrap();
             if segment.value().arguments.is_empty() {
@@ -282,13 +525,14 @@ fn get_argument_modifier(attribute: syn::Attribute) -> Result<ArgumentModifier,
                 match ident {
                     "into" => return Ok(ArgumentModifier::Into),
                     "as_ref" => return Ok(ArgumentModifier::AsRef),
+                    "newtype" => return Ok(ArgumentModifier::Newtype),
                     _ => (),
                 }
             }
         }
     };
 
-    panic!("The attribute argument has to be `from` or `as_ref`, like this: `#[from] a: u32`.")
+    panic!("The attribute argument has to be `into` or `as_ref`, like this: `#[into] a: u32`.")
 }
 
 impl syn::parse::Parse for DelegatedInput {
@@ -301,7 +545,7 @@ impl syn::parse::Parse for DelegatedInput {
             Ok(Self::Argument(expression))
         } else {
             let (input, modifier) = if lookahead.peek(syn::token::Pound) {
-                let mut attributes = input.call(syn::Attribute::parse_outer)?;
+                let mut attributes = input.call(tolerant_outer_attributes)?;
                 if attributes.len() > 1 {
                     panic!("You can specify at most a single attribute for each parameter in a delegated method");
                 }
@@ -323,7 +567,7 @@ impl syn::parse::Parse for DelegatedInput {
 }
 
 struct DelegatedMethod {
-    method: syn::TraitItemMethod,
+    method: syn::TraitItemFn,
     attributes: Vec<syn::Attribute>,
     visibility: syn::Visibility,
     arguments: syn::punctuated::Punctuated<syn::Expr, syn::Token![,]>,
@@ -333,7 +577,7 @@ struct DelegatedMethod {
 // argument used to call the delegate function: omit receiver, extract an
 // identifier from a typed input parameter (and wrap it in an `Expr`).
 fn parse_input_into_argument_expression(
-    function_name: &syn::Ident,
+    function_name: &Ident,
     input: &syn::FnArg,
 ) -> Option<syn::Expr> {
     match input {
@@ -382,7 +626,7 @@ fn parse_input_into_argument_expression(
 
 impl syn::parse::Parse for DelegatedMethod {
     fn parse(input: ParseStream) -> Result<Self, Error> {
-        let attributes = input.call(syn::Attribute::parse_outer)?;
+        let attributes = input.call(tolerant_outer_attributes)?;
         let visibility = input.call(syn::Visibility::parse)?;
 
         // Unchanged from Parse from TraitItemMethod
@@ -391,7 +635,7 @@ impl syn::parse::Parse for DelegatedMethod {
         let unsafety: Option<syn::Token![unsafe]> = input.parse()?;
         let abi: Option<syn::Abi> = input.parse()?;
         let fn_token: syn::Token![fn] = input.parse()?;
-        let ident: syn::Ident = input.parse()?;
+        let ident: Ident = input.parse()?;
         let generics: syn::Generics = input.parse()?;
 
         let content;
@@ -406,8 +650,7 @@ impl syn::parse::Parse for DelegatedMethod {
         // retrieve the expressions inside of the brackets as well as variable
         // identifiers of ordinary inputs. The arguments must preserve the order
         // of the inputs.
-        let delegated_inputs =
-            content.parse_terminated::<DelegatedInput, syn::Token![,]>(DelegatedInput::parse)?;
+        let delegated_inputs = content.parse_terminated(DelegatedInput::parse, syn::Token![,])?;
         let mut inputs: syn::punctuated::Punctuated<syn::FnArg, syn::Token![,]> =
             syn::punctuated::Punctuated::new();
         let mut arguments: syn::punctuated::Punctuated<syn::Expr, syn::Token![,]> =
@@ -472,6 +715,15 @@ impl syn::parse::Parse for DelegatedMethod {
                                 })
                             };
 
+                            let field_call = || {
+                                syn::Expr::from(ExprField {
+                                    attrs: vec![],
+                                    base: Box::new(argument.clone()),
+                                    dot_token: Default::default(),
+                                    member: syn::Member::Unnamed(0.into()),
+                                })
+                            };
+
                             match modifier {
                                 ArgumentModifier::Into => {
                                     argument = method_call("into");
@@ -479,6 +731,7 @@ impl syn::parse::Parse for DelegatedMethod {
                                 ArgumentModifier::AsRef => {
                                     argument = method_call("as_ref");
                                 }
+                                ArgumentModifier::Newtype => argument = field_call(),
                             }
                         }
 
@@ -532,7 +785,7 @@ impl syn::parse::Parse for DelegatedMethod {
         };
 
         // This needs to be populated from scratch because of the signature above.
-        let method = syn::TraitItemMethod {
+        let method = syn::TraitItemFn {
             // All attributes are attached to `DelegatedMethod`, since they
             // presumably pertain to the process of delegation, not the
             // signature of the delegator.
@@ -554,10 +807,14 @@ impl syn::parse::Parse for DelegatedMethod {
 struct DelegatedSegment {
     delegator: syn::Expr,
     methods: Vec<DelegatedMethod>,
+    segment_attrs: SegmentAttributes,
 }
 
 impl syn::parse::Parse for DelegatedSegment {
     fn parse(input: ParseStream) -> Result<Self, Error> {
+        let attributes = input.call(tolerant_outer_attributes)?;
+        let segment_attrs = parse_segment_attributes(&attributes);
+
         if let Ok(keyword) = input.parse::<kw::target>() {
             return Err(Error::new(keyword.span(), "You are using the old `target` expression, which is deprecated. Please replace `target` with `to`."));
         } else {
@@ -570,10 +827,18 @@ impl syn::parse::Parse for DelegatedSegment {
 
             let mut methods = vec![];
             while !content.is_empty() {
-                methods.push(content.parse::<DelegatedMethod>().unwrap());
+                methods.push(
+                    content
+                        .parse::<DelegatedMethod>()
+                        .expect("Cannot parse delegated method"),
+                );
             }
 
-            Ok(DelegatedSegment { delegator, methods })
+            Ok(DelegatedSegment {
+                delegator,
+                methods,
+                segment_attrs,
+            })
         })
     }
 }
@@ -593,188 +858,23 @@ impl syn::parse::Parse for DelegationBlock {
     }
 }
 
-struct CallMethodAttribute {
-    name: syn::Ident,
-}
-
-impl syn::parse::Parse for CallMethodAttribute {
-    fn parse(input: ParseStream) -> Result<Self, Error> {
-        let content;
-        syn::parenthesized!(content in input);
-        Ok(CallMethodAttribute {
-            name: content.parse()?,
-        })
-    }
-}
-
-struct GenerateAwaitAttribute {
-    literal: syn::LitBool,
-}
-
-impl syn::parse::Parse for GenerateAwaitAttribute {
-    fn parse(input: ParseStream) -> Result<Self, Error> {
-        let content;
-        syn::parenthesized!(content in input);
-        Ok(GenerateAwaitAttribute {
-            literal: content.parse()?,
-        })
-    }
-}
-
-enum TryIntoMethod {
-    Raw,
-    Unwrap,
-}
-
-struct ParsedAttributes<'a> {
-    attributes: Vec<&'a syn::Attribute>,
-    target_method: Option<syn::Ident>,
-    generate_into: bool,
-    generate_await: bool,
-    try_into: Option<TryIntoMethod>,
-}
-
-/// Iterates through the attributes of a method and filters special attributes.
-/// - call => sets the name of the target method to call
-/// - into => generates a `into()` call after the delegated expression
-/// - try_into => generates a `try_into()` call after delegated expression
-/// - await => generates an `.await` expression after the delegated expression
-fn parse_attributes<'a>(
-    attrs: &'a [syn::Attribute],
-    method: &syn::TraitItemMethod,
-) -> ParsedAttributes<'a> {
-    let mut target_method: Option<syn::Ident> = None;
-    let mut generate_into: Option<bool> = None;
-    let mut generate_await: Option<bool> = None;
-    let mut try_into: Option<TryIntoMethod> = None;
-
-    // https://github.com/rust-lang/rust/issues/70263
-    #[allow(clippy::type_complexity)]
-    let mut map: HashMap<&str, Box<dyn FnMut(&Attribute)>> = Default::default();
-
-    map.insert(
-        "call",
-        Box::new(|attribute| {
-            let target = syn::parse2::<CallMethodAttribute>(attribute.tokens.clone()).unwrap();
-            if target_method.is_some() {
-                panic!(
-                    "Multiple call attributes specified for {}",
-                    method.sig.ident
-                )
-            }
-            target_method = Some(target.name);
-        }),
-    );
-    map.insert(
-        "target_method",
-        Box::new(|_| {
-            panic!("You are using the old `target_method` attribute, which is deprecated. Please replace `target_method` with `call`.");
-        }),
-    );
-    map.insert(
-        "into",
-        Box::new(|_| {
-            if generate_into.is_some() {
-                panic!(
-                    "Multiple `into` attributes specified for {}",
-                    method.sig.ident
-                )
-            }
-            generate_into = Some(true);
-        }),
-    );
-    map.insert(
-        "try_into",
-        Box::new(|attribute| {
-            if try_into.is_some() {
-                panic!(
-                    "Multiple `try_into` attributes specified for {}",
-                    method.sig.ident
-                )
-            }
-            let meta = attribute
-                .parse_meta()
-                .expect("Invalid `try_into` arguments");
-
-            if let Meta::List(list) = meta {
-                if let Some(syn::NestedMeta::Meta(Meta::Path(path))) = list.nested.first() {
-                    if path.is_ident("unwrap") {
-                        try_into = Some(TryIntoMethod::Unwrap);
-                        return;
-                    }
-                    panic!(
-                        "Unknown `try_into` argument {} specified for {}",
-                        path.to_token_stream(),
-                        method.sig.ident
-                    )
-                }
-            }
-            try_into = Some(TryIntoMethod::Raw);
-        }),
-    );
-    map.insert(
-        "await",
-        Box::new(|attribute| {
-            if generate_await.is_some() {
-                panic!(
-                    "Multiple `await` attributes specified for {}",
-                    method.sig.ident
-                )
-            }
-            let generate = syn::parse2::<GenerateAwaitAttribute>(attribute.tokens.clone()).unwrap();
-            generate_await = Some(generate.literal.value);
-        }),
-    );
-
-    let attrs: Vec<&syn::Attribute> = attrs
-        .iter()
-        .filter(|attr| {
-            if let syn::AttrStyle::Outer = attr.style {
-                for (ident, callback) in map.iter_mut() {
-                    if attr.path.is_ident(ident) {
-                        callback(attr);
-                        return false;
-                    }
-                }
-            }
-
-            true
-        })
-        .collect();
-
-    drop(map);
-
-    ParsedAttributes {
-        attributes: attrs,
-        target_method,
-        generate_into: generate_into.unwrap_or(false),
-        generate_await: generate_await.unwrap_or_else(|| method.sig.asyncness.is_some()),
-        try_into,
-    }
-}
-
 /// Returns true if there are any `inline` attributes in the input.
 fn has_inline_attribute(attrs: &[&syn::Attribute]) -> bool {
     attrs.iter().any(|attr| {
         if let syn::AttrStyle::Outer = attr.style {
-            attr.path.is_ident("inline")
+            attr.path().is_ident("inline")
         } else {
             false
         }
     })
 }
 
-struct MatchVisitor {
-    name: Ident,
-    args: Vec<syn::Expr>,
-}
+struct MatchVisitor<F>(F);
 
-impl VisitMut for MatchVisitor {
+impl<F: Fn(&Expr) -> proc_macro2::TokenStream> VisitMut for MatchVisitor<F> {
     fn visit_arm_mut(&mut self, arm: &mut syn::Arm) {
-        let body = arm.body.clone();
-        let name = &self.name;
-        let args = &self.args;
-        arm.body = parse_quote!( { #body.#name(#(#args),*) });
+        let transformed = self.0(&arm.body);
+        arm.body = parse_quote!(#transformed);
     }
 }
 
@@ -782,14 +882,47 @@ impl VisitMut for MatchVisitor {
 pub fn delegate(tokens: TokenStream) -> TokenStream {
     let block: DelegationBlock = syn::parse_macro_input!(tokens);
     let sections = block.segments.iter().map(|delegator| {
-        let delegator_attribute = &delegator.delegator;
-
+        let delegated_expr = &delegator.delegator;
         let functions = delegator.methods.iter().map(|method| {
             let input = &method.method;
-            let signature = &input.sig;
-
-            let attributes = parse_attributes(&method.attributes, input);
-
+            let mut signature = input.sig.clone();
+            if let Expr::Closure(closure) = delegated_expr {
+                let additional_inputs: Vec<FnArg> = closure
+                    .inputs
+                    .iter()
+                    .map(|input| {
+                        if let syn::Pat::Type(pat_type) = input {
+                            syn::parse_quote!(#pat_type)
+                        } else {
+                            panic!(
+                                "Use a type pattern (`a: u32`) for delegation closure arguments"
+                            );
+                        }
+                    })
+                    .collect();
+                let mut origin_inputs = mem::take(&mut signature.inputs).into_iter();
+                // When delegating methods, `first_input` should be self or similar receivers
+                // Then we need to move it to first
+                // When delegating associated methods, it may be a trivial argument or does not even exist
+                // We just keep the origin order.
+                let first_input = origin_inputs.next();
+                match first_input {
+                    Some(FnArg::Receiver(receiver)) => {
+                        signature.inputs.push(FnArg::Receiver(receiver));
+                        signature.inputs.extend(additional_inputs);
+                    }
+                    Some(first_input) => {
+                        signature.inputs.extend(additional_inputs);
+                        signature.inputs.push(first_input);
+                    }
+                    _ => {
+                        signature.inputs.extend(additional_inputs);
+                    }
+                }
+                signature.inputs.extend(origin_inputs);
+            }
+            let attributes = parse_method_attributes(&method.attributes, input);
+            let attributes = combine_attributes(attributes, &delegator.segment_attrs);
             if input.default.is_some() {
                 panic!(
                     "Do not include implementation of delegated functions ({})",
@@ -798,58 +931,143 @@ pub fn delegate(tokens: TokenStream) -> TokenStream {
             }
 
             // Generate an argument vector from Punctuated list.
-            let args: Vec<syn::Expr> = method.arguments.clone().into_iter().collect();
+            let args: Vec<Expr> = method.arguments.clone().into_iter().collect();
 
-            let name = match &attributes.target_method {
-                Some(n) => n,
-                None => &input.sig.ident,
+            // Get name (or index) of the target method or field
+            let name = match &attributes.target_specifier {
+                Some(target) => target.get_member(&input.sig.ident),
+                None => input.sig.ident.clone().into(),
             };
+
             let inline = if has_inline_attribute(&attributes.attributes) {
                 quote!()
             } else {
-                quote! { #[inline(always)] }
+                quote! { #[inline] }
             };
             let visibility = &method.visibility;
 
-            let body = if let syn::Expr::Match(expr_match) = delegator_attribute {
-                let mut expr_match = expr_match.clone();
-                MatchVisitor {
-                    name: name.clone(),
-                    args,
-                }
-                .visit_expr_match_mut(&mut expr_match);
-                expr_match.into_token_stream()
+            let is_method = method.method.sig.receiver().is_some();
+            let associated_const = &attributes.associated_constant;
+            let expr_attr = &attributes.expr_attr;
+
+            // Use the body of a closure (like `|k: u32| <body>`) as the delegation expression
+            let delegated_body = if let Expr::Closure(closure) = delegated_expr {
+                &closure.body
             } else {
-                quote::quote! { #delegator_attribute.#name(#(#args),*) }
-            };
-            let body = if attributes.generate_await {
-                quote::quote! { #body.await }
-            } else {
-                body
+                delegated_expr
             };
 
             let span = input.span();
-            let body = match &signature.output {
-                syn::ReturnType::Default => quote::quote! { #body; },
-                syn::ReturnType::Type(_, ret_type) => {
-                    if attributes.generate_into {
-                        quote::quote! { ::std::convert::Into::<#ret_type>::into(#body) }
-                    } else if let Some(try_into) = attributes.try_into {
-                        match try_into {
-                            TryIntoMethod::Raw => {
-                                quote::quote! { ::std::convert::TryInto::try_into(#body) }
-                            }
-                            TryIntoMethod::Unwrap => {
-                                quote::quote! { ::std::convert::TryInto::try_into(#body).unwrap() }
-                            }
+            let generate_await = attributes
+                .generate_await
+                .unwrap_or_else(|| method.method.sig.asyncness.is_some());
+
+            // fn method<'a, A, B> -> method::<A, B>
+            let generic_params = &method.method.sig.generics.params;
+            let generics = if generic_params.is_empty() {
+                quote::quote! {}
+            } else {
+                let span = generic_params.span();
+                let mut params: syn::punctuated::Punctuated<
+                    proc_macro2::TokenStream,
+                    syn::Token![,],
+                > = syn::punctuated::Punctuated::new();
+                for param in generic_params.iter() {
+                    let token = match param {
+                        GenericParam::Lifetime(_) => {
+                            // Do not pass lifetimes to generic arguments explicitly to avoid
+                            // things like https://doc.rust-lang.org/error_codes/E0794.html
+                            // See https://github.com/Kobzol/rust-delegate/issues/85.
+                            continue;
                         }
-                    } else {
-                        body
-                    }
+                        GenericParam::Type(t) => {
+                            let token = &t.ident;
+                            let span = t.span();
+                            quote::quote_spanned! {span=> #token }
+                        }
+                        GenericParam::Const(c) => {
+                            let token = &c.ident;
+                            let span = c.span();
+                            quote::quote_spanned! {span=> #token }
+                        }
+                    };
+                    params.push(token);
                 }
+                quote::quote_spanned! {span=> ::<#params> }
             };
 
-            let attrs = attributes.attributes;
+            let modify_expr = |expr: &Expr| {
+                let body = if let Some(target_trait) = &attributes.target_trait {
+                    quote::quote! { #target_trait::#name#generics(#expr, #(#args),*) }
+                } else if let Some(AssociatedConstant {
+                    const_name,
+                    trait_path,
+                }) = associated_const
+                {
+                    let return_type = &signature.output;
+                    quote::quote! {{
+                        const fn get_const<T: #trait_path>(t: &T) #return_type {
+                            <T as #trait_path>::#const_name
+                        }
+                        get_const(#expr)
+                    }}
+                } else if is_method {
+                    match &attributes.target_specifier {
+                        None | Some(TargetSpecifier::Method(_)) => {
+                            quote::quote! { #expr.#name#generics(#(#args),*) }
+                        }
+                        Some(TargetSpecifier::Field(target)) => {
+                            let reference = target.reference_tokens();
+                            quote::quote! { #reference#expr.#name }
+                        }
+                    }
+                } else {
+                    quote::quote! { #expr::#name#generics(#(#args),*) }
+                };
+
+                let mut body = if generate_await {
+                    quote::quote! { #body.await }
+                } else {
+                    body
+                };
+
+                for expression in &attributes.expressions {
+                    match expression {
+                        ReturnExpression::Into(type_name) => {
+                            body = match type_name {
+                                Some(name) => {
+                                    quote::quote! { ::core::convert::Into::<#name>::into(#body) }
+                                }
+                                None => quote::quote! { ::core::convert::Into::into(#body) },
+                            };
+                        }
+                        ReturnExpression::TryInto => {
+                            body = quote::quote! { ::core::convert::TryInto::try_into(#body) };
+                        }
+                        ReturnExpression::Unwrap => {
+                            body = quote::quote! { #body.unwrap() };
+                        }
+                    }
+                }
+                body
+            };
+            let mut body = if let Expr::Match(expr_match) = delegated_body {
+                let mut expr_match = expr_match.clone();
+                MatchVisitor(modify_expr).visit_expr_match_mut(&mut expr_match);
+                expr_match.into_token_stream()
+            } else {
+                modify_expr(delegated_body)
+            };
+
+            if let syn::ReturnType::Default = &signature.output {
+                body = quote::quote! { #body; };
+            };
+
+            if let Some(expr_template) = expr_attr {
+                body = expr_template.expand_template(&body);
+            }
+
+            let attrs = &attributes.attributes;
             quote::quote_spanned! {span=>
                 #(#attrs)*
                 #inline
@@ -867,3 +1085,93 @@ pub fn delegate(tokens: TokenStream) -> TokenStream {
     };
     result.into()
 }
+
+// we cannot use `Attributes::parse_outer` directly, because it does not allow keywords to appear
+// in meta path positions, i.e., it does not accept `#[await(true)]`.
+// related issue: https://github.com/dtolnay/syn/issues/1458
+fn tolerant_outer_attributes(input: ParseStream) -> syn::Result<Vec<syn::Attribute>> {
+    use proc_macro2::{Delimiter, TokenTree};
+    use syn::{
+        bracketed,
+        ext::IdentExt,
+        parse::discouraged::Speculative,
+        token::{Brace, Bracket, Paren},
+        AttrStyle, Attribute, ExprLit, Lit, MacroDelimiter, MetaList, MetaNameValue, Path, Result,
+        Token,
+    };
+
+    fn tolerant_attr(input: ParseStream) -> Result<Attribute> {
+        let content;
+        Ok(Attribute {
+            pound_token: input.parse()?,
+            style: AttrStyle::Outer,
+            bracket_token: bracketed!(content in input),
+            meta: content.call(tolerant_meta)?,
+        })
+    }
+
+    // adapted from `impl Parse for Meta`
+    fn tolerant_meta(input: ParseStream) -> Result<Meta> {
+        // Try to parse as Meta
+        if let Ok(meta) = input.call(Meta::parse) {
+            Ok(meta)
+        } else {
+            // If it's not possible, try to parse it as any identifier, to support #[await]
+            let path = Path::from(input.call(Ident::parse_any)?);
+            if input.peek(Paren) || input.peek(Bracket) || input.peek(Brace) {
+                // adapted from the private `syn::attr::parse_meta_after_path`
+                input.step(|cursor| {
+                    if let Some((TokenTree::Group(g), rest)) = cursor.token_tree() {
+                        let span = g.delim_span();
+                        let delimiter = match g.delimiter() {
+                            Delimiter::Parenthesis => MacroDelimiter::Paren(Paren(span)),
+                            Delimiter::Brace => MacroDelimiter::Brace(Brace(span)),
+                            Delimiter::Bracket => MacroDelimiter::Bracket(Bracket(span)),
+                            Delimiter::None => {
+                                return Err(cursor.error("expected delimiter"));
+                            }
+                        };
+                        Ok((
+                            Meta::List(MetaList {
+                                path,
+                                delimiter,
+                                tokens: g.stream(),
+                            }),
+                            rest,
+                        ))
+                    } else {
+                        Err(cursor.error("expected delimiter"))
+                    }
+                })
+            } else if input.peek(Token![=]) {
+                // adapted from the private `syn::attr::parse_meta_name_value_after_path`
+                let eq_token = input.parse()?;
+                let ahead = input.fork();
+                let value = match ahead.parse::<Option<Lit>>()? {
+                    // this branch is probably for speeding up the parsing for doc comments etc.
+                    Some(lit) if ahead.is_empty() => {
+                        input.advance_to(&ahead);
+                        Expr::Lit(ExprLit {
+                            attrs: Vec::new(),
+                            lit,
+                        })
+                    }
+                    _ => input.parse()?,
+                };
+                Ok(Meta::NameValue(MetaNameValue {
+                    path,
+                    eq_token,
+                    value,
+                }))
+            } else {
+                Ok(Meta::Path(path))
+            }
+        }
+    }
+
+    let mut attrs = Vec::new();
+    while input.peek(Token![#]) {
+        attrs.push(input.call(tolerant_attr)?);
+    }
+    Ok(attrs)
+}
